/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif

#ifdef UBASIC_TEST
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#endif

#include "tokenizer.h"
#include "stdlib.h"

static char const *ptr, *first_ptr, *hlptr, *nextptr=0;
char statement[128]; // debugstring
static char tok[4]; //

#define MAX_VARNUM 64 // 2*26 'a'-'Z' +#0-#9 +## + #@=loop 
#define MAX_NUMLEN 9  //:...i....1....;....2....;....3....;....4..

struct keyword_token {
  const char *keyword;
  short token;
};

static ubasic_token current_token = TOKENIZER_ERROR;
static int   current_line = 1;
static int   first_line=1;
static short sep_memtoken = 0;
 char debug = 0;
 
static const struct keyword_token keywords[] =
 {  // lowercase ONLY!!
  {"not",                     TOKENIZER_LNOT},
  {"or",                      TOKENIZER_LOR},
  {"and",                     TOKENIZER_LAND},
  {"mod",                      TOKENIZER_MOD},
  {"xor",                      TOKENIZER_XOR},
  {"abs",                     TOKENIZER_ABS}, // // -x -->> +x
  {"odd",                      TOKENIZER_ODD}, // // 0 2 4 6 8 => 1

  {"if",                      TOKENIZER_IF},
  {"then",                    TOKENIZER_THEN},
  {"elseif",                    TOKENIZER_ELSEIF},
  {"else",                    TOKENIZER_ELSE},
  {"endif",                   TOKENIZER_ENDIF},
  {"select",                  TOKENIZER_SELECT},
  {"case_else",               TOKENIZER_CASE_ELSE},
  {"case",                    TOKENIZER_CASE}, // autobreak
  {"end_select",              TOKENIZER_END_SELECT},
  {"exec",                    TOKENIZER_EXEC},
  {"for",                     TOKENIZER_FOR},
  {"to",                      TOKENIZER_TO},
  {"next",                    TOKENIZER_NEXT},
  {"step",                    TOKENIZER_STEP},
  {"do",                      TOKENIZER_DO}, // ++ bubble behind then, else, case
  {"until",                   TOKENIZER_UNTIL},
  {"while",                   TOKENIZER_WHILE},
  {"wend",                    TOKENIZER_WEND},
  {"break",               TOKENIZER_BREAK}, // // do for while & see data statement;....4..
  {"cont",                TOKENIZER_CONT}, // // jump to -> wend, until, next
  {"goto",                    TOKENIZER_GOTO}, // ++ calculated jump method: goto 100*i+x 
  {"gosub",                   TOKENIZER_GOSUB}, //++ gosub 100*i+x 
  {"return",                  TOKENIZER_RETURN},
  {"skip",                      TOKENIZER_SKIP}, // >> Cracy Jump >> token
  {"data",                      TOKENIZER_DATA}, //feeds var #0,..,..,#9
  {"let",                     TOKENIZER_LET}, // speeds up on 1. Position
  {"rem",                     TOKENIZER_REM},
  {"cls",                     TOKENIZER_CLS},
  {"print_screen",            TOKENIZER_PRINT_SCREEN},
  {"print",                   TOKENIZER_PRINT},
  {"beep",					TOKENIZER_BEEP}, //:...i....1....;....2....;....3....;....4..
  {"playsound",               TOKENIZER_PLAY_SOUND},
  {"is_key",                  TOKENIZER_IS_KEY}, // ++ is_key K "up" 100+x -->> K=100+x
  {"is_pressed",              TOKENIZER_IS_PRESSED}, // ++ is_pressed K "up" 100*x -->> gosub K data ....,...,...
  {"click",                   TOKENIZER_CLICK},
  {"press",                   TOKENIZER_PRESS},
  {"release",                 TOKENIZER_RELEASE},
  //{"shot",                    TOKENIZER_SHOOT}, // for compatibility
  {"shoot",                   TOKENIZER_SHOOT}, // ++ shoot ~ // rawproof?
  {"md_detect_motion",        TOKENIZER_MD_DETECT_MOTION},

  /* WARNING due to tokenizer limitation longest match must be first */
// GET  
  {"get_shooting",            TOKENIZER_GET_SHOOTING},
  {"get_av96",                TOKENIZER_GET_AV96},
  {"get_av",                  TOKENIZER_GET_USER_AV_ID}, //FOR COMPATIBILITY
  {"get_bv96",                TOKENIZER_GET_BV96},
  {"get_capture_mode",        TOKENIZER_GET_CAPTURE_MODE},
  {"get_config_value",        TOKENIZER_GET_CONFIG_VALUE},
  {"get_display_mode",        TOKENIZER_GET_DISPLAY_MODE},
  {"get_day_seconds",         TOKENIZER_GET_DAY_SECONDS},
  {"get_disk_size",           TOKENIZER_GET_DISK_SIZE},
  {"get_dof",                 TOKENIZER_GET_DOF},
  {"get_draw_title_line",     TOKENIZER_GET_DRAW_TITLE_LINE},
  {"get_drive_mode",          TOKENIZER_GET_DRIVE_MODE},
  {"get_exp_count",           TOKENIZER_GET_EXP_COUNT},
  {"get_flash_mode",          TOKENIZER_GET_FLASH_MODE},
  {"get_flash_ready",         TOKENIZER_GET_FLASH_READY},
  {"get_far_limit",           TOKENIZER_GET_FAR_LIMIT},
  {"get_free_disk_space",     TOKENIZER_GET_FREE_DISK_SPACE},
  {"get_focal_length",        TOKENIZER_GET_FOCAL_LENGTH},
  {"get_focus_mode",          TOKENIZER_GET_FOCUS_MODE},
  {"get_focus_ok",            TOKENIZER_GET_FOCUS_OK},
  {"get_focus_state",         TOKENIZER_GET_FOCUS_STATE},
  {"get_focus",               TOKENIZER_GET_FOCUS},
  {"get_sd_over_modes",       TOKENIZER_GET_SD_OVER_MODES},
  {"get_hyp_dist",            TOKENIZER_GET_HYPERFOCAL_DIST},
  {"get_iso_market",          TOKENIZER_GET_ISO_MARKET},
  {"get_iso_mode",            TOKENIZER_GET_ISO_MODE},
  {"get_iso_real",            TOKENIZER_GET_ISO_REAL},
  {"get_iso",                 TOKENIZER_GET_ISO_MODE}, //FOR COMPATIBILITY
  {"get_is_mode",             TOKENIZER_GET_IS_MODE}, // =get_IS_mode ; upper|lower IGNORED NOW in script
  {"get_jpg_count",           TOKENIZER_GET_JPG_COUNT},
  {"get_min_stack_dist",      TOKENIZER_GET_MIN_STACK_DIST},
  {"get_movie_status",        TOKENIZER_GET_MOVIE_STATUS},
  {"get_mode",				  TOKENIZER_GET_MODE},  // Returns 0 in recordmode, 1 in playmode
  {"get_near_limit",          TOKENIZER_GET_NEAR_LIMIT},
  {"get_platform_id",         TOKENIZER_GET_PLATFORM_ID},
  {"get_propset",             TOKENIZER_GET_PROPSET},
  {"get_prop",                TOKENIZER_GET_PROP},
  {"get_quality",             TOKENIZER_GET_QUALITY},
  {"get_raw_support",         TOKENIZER_GET_RAW_SUPPORT},
  {"get_raw_count",           TOKENIZER_GET_RAW_COUNT},
  {"get_raw_nr",              TOKENIZER_GET_RAW_NR},
  {"get_raw",                 TOKENIZER_GET_RAW},
  {"get_resolution",          TOKENIZER_GET_RESOLUTION},
  {"get_sv96",	              TOKENIZER_GET_SV96},
  {"get_temperature",         TOKENIZER_GET_TEMPERATURE},
  {"get_tick_count",          TOKENIZER_GET_TICK_COUNT},
  {"gtc",          			TOKENIZER_GET_TICK_COUNT},
  {"get_time",          	  TOKENIZER_GET_TIME},
  {"get_tv96",                TOKENIZER_GET_TV96},
  {"get_user_av_id",          TOKENIZER_GET_USER_AV_ID},
  {"get_user_av96",           TOKENIZER_GET_USER_AV96},
  {"get_user_tv_id",          TOKENIZER_GET_USER_TV_ID},
  {"get_user_tv96",           TOKENIZER_GET_USER_TV96},
  {"get_video_recording",     TOKENIZER_GET_VIDEO_RECORDING},
  {"get_video_button",        TOKENIZER_GET_VIDEO_BUTTON},
  {"get_vbatt",               TOKENIZER_GET_VBATT},  
  {"get_zoom_steps",          TOKENIZER_GET_ZOOM_STEPS},
  {"get_zoom",                TOKENIZER_GET_ZOOM},
  {"is_capture_mode_valid",   TOKENIZER_IS_CAPTURE_MODE_VALID},
//SET
  {"set_aelock",              TOKENIZER_SET_AELOCK},
  {"set_aflock",              TOKENIZER_SET_AFLOCK},
  {"set_av96_direct",         TOKENIZER_SET_AV96_DIRECT},
  {"set_av96",                TOKENIZER_SET_AV96},
  {"set_av_rel",              TOKENIZER_SET_USER_AV_BY_ID_REL}, //FOR COMPATIBILITY
  {"set_av",                  TOKENIZER_SET_USER_AV_BY_ID}, //FOR COMPATIBILITY
  {"set_backlight",           TOKENIZER_SET_BACKLIGHT},
  {"set_capture_mode_canon",  TOKENIZER_SET_CAPTURE_MODE_CANON},
  {"set_capture_mode",        TOKENIZER_SET_CAPTURE_MODE},
  {"set_config_value",        TOKENIZER_SET_CONFIG_VALUE},
  {"set_draw_title_line",     TOKENIZER_SET_DRAW_TITLE_LINE},
  {"set_exit_key",            TOKENIZER_SET_EXIT_KEY},
  {"set_focus",               TOKENIZER_SET_FOCUS},
  {"set_iso_mode",            TOKENIZER_SET_ISO_MODE},
  {"set_iso_real",            TOKENIZER_SET_ISO_REAL},
  {"set_iso",                 TOKENIZER_SET_ISO_MODE}, //FOR COMPATIBILITY
  {"set_lcd_display",         TOKENIZER_SET_LCDDISPLAY},
  {"set_led",                 TOKENIZER_SET_LED},
  {"set_mf",                  TOKENIZER_SET_MF},
  {"set_movie_status",        TOKENIZER_SET_MOVIE_STATUS},  
  {"set_nd_filter",           TOKENIZER_SET_ND_FILTER},
  {"set_prop",                TOKENIZER_SET_PROP},
  {"set_quality",             TOKENIZER_SET_QUALITY},
  {"set_raw_nr",              TOKENIZER_SET_RAW_NR},
  {"set_raw",                 TOKENIZER_SET_RAW},
  {"set_record",              TOKENIZER_SET_RECORD},
  {"set_remote_timing",       TOKENIZER_SET_REMOTE_TIMING},
  {"set_resolution",          TOKENIZER_SET_RESOLUTION},
  {"set_sv96",		          TOKENIZER_SET_SV96},
  {"set_tv96_direct",         TOKENIZER_SET_TV96_DIRECT},
  {"set_tv96",                TOKENIZER_SET_TV96},
  {"set_tv_rel",              TOKENIZER_SET_USER_TV_BY_ID_REL}, //FOR COMPATIBILITY
  {"set_tv",                  TOKENIZER_SET_USER_TV_BY_ID}, //FOR COMPATIBILITY
  {"set_user_av_by_id_rel",   TOKENIZER_SET_USER_AV_BY_ID_REL},
  {"set_user_av_by_id",       TOKENIZER_SET_USER_AV_BY_ID},
  {"set_user_av96",           TOKENIZER_SET_USER_AV96},
  {"set_user_tv_by_id_rel",   TOKENIZER_SET_USER_TV_BY_ID_REL},
  {"set_user_tv_by_id",       TOKENIZER_SET_USER_TV_BY_ID},
  {"set_user_tv96",           TOKENIZER_SET_USER_TV96},
  {"set_yield",               TOKENIZER_SET_YIELD},
  {"set_zoom_speed",          TOKENIZER_SET_ZOOM_SPEED},
  {"set_zoom_rel",            TOKENIZER_SET_ZOOM_REL},
  {"set_zoom",                TOKENIZER_SET_ZOOM},
  
  {"wait_click",              TOKENIZER_WAIT_CLICK},
  {"wheel_right",             TOKENIZER_WHEEL_RIGHT},
  {"wheel_left",              TOKENIZER_WHEEL_LEFT},
  /*
  {"@debug",                  TOKENIZER_REM}, // opt. 1st-3rd! line to build a string, slows down a bit
  {"@title",                  TOKENIZER_REM},
  {"@subtitle",               TOKENIZER_REM},
  {"@param",                  TOKENIZER_REM},
  {"@default",                TOKENIZER_REM},
  {"@range",                  TOKENIZER_REM},
  {"@values",                 TOKENIZER_REM},
  {"@chdk_version",           TOKENIZER_REM},  //:...i....1....;....2....;....3....;....4..*/

  {"md_get_cell_diff",        TOKENIZER_MD_GET_CELL_DIFF},
  {"md_get_cell_val",         TOKENIZER_MD_GET_CELL_VAL},
  {"md_af_led_control",       TOKENIZER_MD_AF_LED_CONTROL},
  {"exit_alt",                TOKENIZER_EXIT_ALT},
  {"enter_alt",               TOKENIZER_ENTER_ALT}, 
  {"get_alt_mode",            TOKENIZER_GET_ALT_MODE}, 
  
  {"set_ev",                  TOKENIZER_SET_EV},
  {"get_ev",                  TOKENIZER_GET_EV},
  {"get_orientation_sensor",  TOKENIZER_GET_ORIENTATION_SENSOR},
  {"get_nd_present",          TOKENIZER_GET_ND_PRESENT},
  {"get_histo_range",         TOKENIZER_GET_HISTO_RANGE},
  {"autostarted",             TOKENIZER_SCRIPT_AUTOSTARTED},
  {"get_autostart",           TOKENIZER_GET_SCRIPT_AUTOSTART},
  {"set_autostart",           TOKENIZER_SET_SCRIPT_AUTOSTART},
  {"get_usb_power",           TOKENIZER_GET_USB_POWER},
  {"usb_force_active",        TOKENIZER_FORCE_USB_PRESENT},
  {"usb_sync_wait",           TOKENIZER_USB_SYNC_WAIT}, 
  {"shot_histo_enable",       TOKENIZER_SHOT_HISTO_ENABLE},
  {"set_console_layout",      TOKENIZER_SET_CONSOLE_LAYOUT},
  {"set_console_autoredraw",  TOKENIZER_SET_CONSOLE_AUTOREDRAW},
  {"console_redraw",          TOKENIZER_CONSOLE_REDRAW},
  {"swap_partitions",         TOKENIZER_SWAP_PARTITIONS},
  {"sleep",                   TOKENIZER_SLEEP}, 
  {"random",                  TOKENIZER_RANDOM},
  {"reboot",                  TOKENIZER_REBOOT},
  {"shut_down",               TOKENIZER_SHUT_DOWN},
  
  // APEX functions
  {"iso_to_sv96",             TOKENIZER_ISO_TO_SV96},
  {"sv96_to_iso",             TOKENIZER_SV96_TO_ISO},
  {"iso_real_to_market",      TOKENIZER_ISO_REAL_TO_MARKET},
  {"iso_market_to_real",      TOKENIZER_ISO_MARKET_TO_REAL},
  {"sv96_real_to_market",     TOKENIZER_SV96_REAL_TO_MARKET},
  {"sv96_market_to_real",     TOKENIZER_SV96_MARKET_TO_REAL},
  {"aperture_to_av96",        TOKENIZER_APERTURE_TO_AV96},
  {"av96_to_aperture",        TOKENIZER_AV96_TO_APERTURE},
  {"usec_to_tv96",            TOKENIZER_USEC_TO_TV96},
  {"tv96_to_usec",            TOKENIZER_TV96_TO_USEC},
  {"seconds_to_tv96",         TOKENIZER_SECONDS_TO_TV96},
  
   //:...i....1....;....2....;....3....;....4..
  {"varlist",               TOKENIZER_VARLIST}, // vars<>0
  {"lsec",        		TOKENIZER_LSEC}, // linetime,looptime
  {"msec",        		TOKENIZER_MSEC}, // since Start:
  {"sec",        		TOKENIZER_SEKUNDE},//
  {"min",  				TOKENIZER_MINUTE},//
  {"hour",  				TOKENIZER_STUNDE},//
  {"true",  				TOKENIZER_TRUE}, // !0
  {"false",  			  TOKENIZER_FALSE}, // 0
  {"off",  				  TOKENIZER_OFF}, // 0
  {"on",  				  TOKENIZER_ON}, // 1
  {"swap",  				TOKENIZER_SWAP}, // i,j ; k,l ; #7,m...
  {"reset",               TOKENIZER_RESET},
  {"loop",               TOKENIZER_LOOP}, // 1 = restart if error|eof // |for i=0|1,loop ...
  {"list",               TOKENIZER_LIST}, // Line prefix: show & do
  {"tronly",			  TOKENIZER_TRONLY}, // 'list' !do >>
  {"tron",				  TOKENIZER_TRON}, // 'list' >>
  {"troff",				  TOKENIZER_TROFF},// <<
  {"bug",               TOKENIZER_BUG}, // preprogrammable Error // ?bug,bug gives linenr 
  {"debug",               TOKENIZER_DEBUG}, 
  {"clean",                     TOKENIZER_CLEAN}, 
  
  {"end",                     TOKENIZER_END},  // ++ end X :show console X%10 sec.
  {NULL,                      TOKENIZER_ERROR}
};

/*---------------------------------------------------------------------------*/
static short singlechar(void)
{
  if(*ptr == '\r') ptr++;
  if(*ptr == '\n') {
   strcat(statement,"@");	  return TOKENIZER_CR;
  } else if(*ptr == '@') { tokenizer_eol(); return TOKENIZER_CR;
  } else if(*ptr == '#') {    return TOKENIZER_VARIABLE;
  } else if(*ptr == '=') {    return TOKENIZER_EQ;
  } else if(*ptr == '<') //:...i....1....;....2....;....3....;....4..
	   { if(*(ptr+1) == '=')  { ptr++; return TOKENIZER_LE;
	} else if(*(ptr+1)=='>')  { ptr++; return TOKENIZER_NE; }
		 else return TOKENIZER_LT;
  } else if(*ptr == '>') 
	   { if(*(ptr+1) == '=')  { ptr++; return TOKENIZER_GE; } // "<>" == "!=" | "~="
		 else return TOKENIZER_GT;
  } else if(*ptr == '!') 
	   { if(*(ptr+1) == '=')  { ptr++; return TOKENIZER_NE; }
		 else return TOKENIZER_LNOT;
  } else if(*ptr == '~') {    return TOKENIZER_SEP;  // '~' allows multi statement lines
  } else if(*ptr == ',') {    return TOKENIZER_COMMA;
  } else if(*ptr == '+') {    return TOKENIZER_PLUS;
  } else if(*ptr == '-') {    return TOKENIZER_MINUS;
  } else if(*ptr == '*') {    return TOKENIZER_ASTR;
  } else if(*ptr == '/') {    return TOKENIZER_SLASH;
  } else if(*ptr == '%') {    return TOKENIZER_MOD;
  } else if(*ptr == '&') {    return TOKENIZER_AND;
  } else if(*ptr == '|') {    return TOKENIZER_OR;
  } else if(*ptr == '^') {    return TOKENIZER_XOR;
  } else if(*ptr == '(') {    return TOKENIZER_LEFTPAREN;
  } else if(*ptr == ')') {    return TOKENIZER_RIGHTPAREN;
  } else if(*ptr == ';') {    return TOKENIZER_SEMICOLON;
  } else if(*ptr == '?') {    return TOKENIZER_PRINT;  // Caefix, 6.4.19
  } else if(*ptr == '\\'){ // while (abs(r)>9) {r=r/10;}; '\-4210'=-4
	 while (*(ptr+1) == '\\') ptr++; return TOKENIZER_CAP; // avoid RegExp ->.Log
  } else if(*ptr == '\''){    return TOKENIZER_SEKUNDE; // a sleep 1000
  } else if(*ptr == '.') {    return TOKENIZER_MOMENT; // a sleep (50)
  } else if(*ptr == '') {    return TOKENIZER_BEEP; // Major update! // Yes, Sir! // Wheres the joke?
  } else if(*ptr == ' ') {    return TOKENIZER_ERROR;
//  } else if(*ptr == '"') {    return TOKENIZER_STRING;
//  } else if(*ptr == ':') {    return TOKENIZER_LABEL;
/*
  } else if(isdigit(*ptr)){ return TOKENIZER_NUMBER;*/
  }
  //return 0;
  return TOKENIZER_BUG;
}
/*---------------------------------------------------------------------------*/
static void wskip(void)  // eat all {whitespace}
{
while(*ptr == ' ' || *ptr == '\t'  || *ptr == '}') ptr++;
  if (*ptr == '{') {//3....,....4.. rem exec
	do {ptr++;
		if (*ptr =='\n') current_line++;
	}while (*ptr != '}' && *ptr!=0);
	wskip(); };
  if (*ptr == '@') {tokenizer_eol();}
//  if (*ptr =='\n') {current_token = TOKENIZER_CR;} else
//  if (*ptr ==  0 ) {current_token = TOKENIZER_ENDOFINPUT;}
nextptr=ptr;
}
/*---------------------------------------------------------------------------*/ 
 // Caefix, :...,....1....,....2....,....3....,....4..
void tokenizer_eol(void)
{
	if (current_token != TOKENIZER_CR)
    ptr=nextptr;
	while (*ptr != '\n' && *ptr != 0) {++ptr;}
	if (*ptr == '\n') {current_token=TOKENIZER_CR;}
	else {current_token=TOKENIZER_ENDOFINPUT;}
  nextptr=ptr;
}
/* 
*/ 
void tokenizer_inc(void)
{ 
  ptr=nextptr;
	current_token=singlechar();
	if (current_token==TOKENIZER_BUG) ptr--;
  nextptr=ptr+1;
}
/* 
*/ 
void tokenizer_sep(void)
{	
//	if (*ptr == '\n') return;
	sep_memtoken=current_token;
	current_token=TOKENIZER_SEP;
	//current_token=singlechar();
	
}

/*---------------------------------------------------------------------------*/
	// if ( t=tokenizer_char() != TOKENIZER_BUG)
short tokenizer_char(void)
{ 
	if (current_token == TOKENIZER_VARIABLE) return TOKENIZER_VARIABLE;
	return singlechar();
}
/*---------------------------------------------------------------------------*/
static short get_next_token(void)
{
  struct keyword_token const *kt;
  short  i,x, t=TOKENIZER_ENDOFINPUT;

  DEBUG_PRINTF("get_next_token(): '%s'\n", ptr);
  
  if (*ptr == 0) { return t; }
    // UnknStatement should have size, otherwise hanging-up in ubasic.c possible for some cases 
  nextptr = ptr + 1; 
  if (*ptr=='[') { t=abs(atoi(nextptr));
	if (!t||t>TOKENIZER_END) return TOKENIZER_BUG; else ptr++;} 	// save s, count [token] in tokenizer.h //
  if (isdigit(*ptr)) {
    for (i = 0; i <= (MAX_NUMLEN); ++i) {
      if (!isdigit(ptr[i])) {
	if (i > 0) {
	  nextptr = ptr + i;
      if (*nextptr==']') { ptr--; nextptr++; 	// [3]A ist noch schneller als #A //
		if (debug) { tok[0]='\0';
			x=strlen(statement);
			if (x<120) {
			sprintf(tok, "%d%d%d",t/100,t%100/10,t%10);
			statement[x]='[';
			statement[x+1]='\0';
			strcat(statement,tok);
			statement[x+4]=']';
			statement[x+5]='\0';}
		} return t;}
	  if (debug) { x=strlen(statement);
		//if (x<127) 
		{statement[x]=ptr[0]; statement[x+1]='\0';}
		}
	  return TOKENIZER_NUMBER;
	} else {
	  DEBUG_PRINTF("get_next_token: error due to too short number\n");
	  return TOKENIZER_ERROR;
	}
      }
      if(!isdigit(ptr[i])) {
	DEBUG_PRINTF("get_next_token: error due to malformed number\n");
	return TOKENIZER_ERROR;
      }
    }
    DEBUG_PRINTF("get_next_token: error due to too long number\n");
    return TOKENIZER_ERROR;
	} else if(*ptr==':') {
		if (current_token != TOKENIZER_CR)
		{nextptr = ptr + 1; return TOKENIZER_LABEL;} //  ?"line 1",a:"line 2=";b,c:"line 3".
    // label
      nextptr = ptr;
      do {++nextptr;}
	  while(*nextptr != ' ' && *nextptr != '\r' && *nextptr != '\n'
		 && *nextptr != '\t' && *nextptr != '\0');
	  return TOKENIZER_LABEL;
	} else if(*ptr=='"') {
      nextptr = ptr;
      do { ++nextptr; }
	  while(*nextptr != '"' && *nextptr != '\r' && *nextptr != '\n'
		 && *nextptr != '\0'); // avoid open string error
      while(*nextptr == '"') ++nextptr;
	  return TOKENIZER_STRING;
	
  } else if((i=singlechar()) != TOKENIZER_BUG) { 
	 // if (i == TOKENIZER_ERROR) {ptr++; i=singlechar()}
	  if (i == TOKENIZER_CR) { nextptr = ptr;
    /*
	// move to next line, and skip all following empty lines as well
      while (*ptr == '\n') 
      {
        current_line++;
        ptr++; wskip();
        // eat all whitespace
       // while(*ptr == ' ' || *ptr == '\t' || *ptr == '\r') ptr++;
      };
      ptr--;
      // dangelo: now the last char might point to a whitespace instead of
      // a CR. I hope that doesn't break anything.
	  */
	} else {
      nextptr = ptr + 1;
	//  if (debug)
	  {
		x=strlen(statement);
		//if (x<63) 
		{statement[x]=ptr[0]; statement[x+1]='\0';}
	  }
	}
    return i;
	
  } else {
    for (kt = keywords; kt->keyword != NULL; ++kt)
	{   x=strlen(kt->keyword);
     // x=(//strncmp(ptr, kt->keyword, 2) == 0 &&
      //   strncmp(ptr, kt->keyword, x) == 0)*x;
		// *:...;....1....;....2....;
		hlptr=kt->keyword; i=0;
		while ( i<x ) {
			if (*(ptr+i) == '.') {if(i+1<x) x=i+1; break;} // ++ Autovervollstndigung: 'Wai. 0' // 1. Treffer in Liste!!!
			if ( ((*(ptr+i) | (*(ptr+i)==95?95:32)) != *(hlptr+i)) ) //I  ++ Ignore_CASE  
				{x=0; break;}
			++i;
		}	 ///*/
		//:...;....1....;....2....;
		if (x) {
		nextptr = ptr + x;
		if (current_token==TOKENIZER_CR)
			strcpy(statement,hlptr);
		else if (debug>1) {
		   if (x<8) strcat(statement,hlptr); else { // [
			x=strlen(statement);
			if (x<120) { tok[0]='\0'; t= kt->token&255;
				sprintf(tok, "%d%d%d",t/100,t%100/10,t%10);
				statement[x]='[';
				statement[x+1]='\0';
				strcat(statement,tok);
				statement[x+4]=']';
				statement[x+5]='\0';} } // ]
			} else strcat(statement,"_");
        if (kt->token == TOKENIZER_REM) {
			if (*nextptr=='~')  // old uBasic ignores, new not.
				{nextptr++; return TOKENIZER_SEP;}
			tokenizer_eol(); return TOKENIZER_CR;}
		  // { while(*nextptr != 0 && *nextptr != '\r' && *nextptr != '\n') ++nextptr;}
		return kt->token;
		} // if x 
    } // for kt
  }

  if((*ptr >= 'a' && *ptr <= 'z') || (*ptr >= 'A' && *ptr <= 'Z')) {
    nextptr = ptr + 1;
	
	if (debug>1) 
		{x=strlen(statement);
		//if (x<64) 
		{statement[x]=ptr[0]; statement[x+1]='\0';}
	}
    return TOKENIZER_VARIABLE;
  }
  
  return TOKENIZER_ERROR;
}
/*---------------------------------------------------------------------------*/
short tokenizer_token(void)
{
  return current_token;
}
/*---------------------------------------------------------------------------*/
void tokenizer_next(void)
{
  while (sep_memtoken) {ptr = nextptr;
	  current_token = sep_memtoken;
	  sep_memtoken =0;
	  if (current_token == TOKENIZER_CR) break;
	  return;}
  
  if (current_token == TOKENIZER_CR) {
	statement[0]='\0';
	do {    // move to next line, and skip all following empty lines as well
		ptr++; current_line++;
		wskip();
	}while (*ptr == '\n');
  
 } else { ptr = nextptr; wskip(); }
  
        DEBUG_PRINTF("tokenizer_next: %p\n", nextptr);
  current_token = get_next_token();
        DEBUG_PRINTF("tokenizer_next: '%s' %d\n", ptr, current_token);
  return;
}
/*---------------------------------------------------------------------------*/
void tokenizer_exec(void)  /// switch multilinecomment  {to function		}
{  /// then exec { else "ignore" {todo block
while(*ptr != '{' &&  *ptr != '\n' && *ptr != '\0') ptr++;
  if (*ptr == '{') ptr++; ///{3....,....4.. first '}' closes All '{' !!
  nextptr=ptr;
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
void tokenizer_skip(short token)
{
	while (current_token != token &&
	current_token != TOKENIZER_ENDOFINPUT)
	{tokenizer_next();}
}
/*---------------------------------------------------------------------------*/
void tokenizer_init(const char *program)
{  
  ptr = program;
  if (current_line==1) { // no need to read "twice", nice. Caefix
    memcpy(statement, ptr+1, 127);
	statement[127]='\0';
	hlptr=strpbrk(statement,"debug");
	if (hlptr>ptr) {debug=atoi(hlptr+1);
		if (debug<0) debug=1;}
	tok[3]='0'; 
	nextptr=ptr+1;
	tokenizer_eol();
	tokenizer_next();
	nextptr=ptr-1;
	first_line=current_line-1;
	first_ptr=ptr-1;
	}
  
  statement[0]='\0';
  current_line=first_line;
  current_token=TOKENIZER_CR;
  nextptr=first_ptr;
  ptr=first_ptr;
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
int tokenizer_line_number(void)
{
  return current_line;
}
/*---------------------------------------------------------------------------*/
long tokenizer_value(void)
{
  return atoi(ptr);
}
/*---------------------------------------------------------------------------*/
short tokenizer_variable_num(void)
{
long r=0;
 
 if (*ptr=='#') {
	 ptr++; nextptr=ptr+1;
	 if (*ptr=='#') {return MAX_VARNUM-2;};
	 if (*ptr=='@') {
		ptr++; nextptr++; return MAX_VARNUM-1;}; // == loop
	 if (*ptr=='(') // NOT : let #(i+x)=... // but x=|data #(i+x)..., // varfactor only &63 CAREful
		{current_token=TOKENIZER_LEFTPAREN; return MAX_VARNUM-2;}
	 if (isdigit(*ptr)) {
		return tokenizer_value()%10+(MAX_VARNUM-12); };
	};
 r = *ptr - (*ptr>='a'?'a':('A'-26));
 if (r>=MAX_VARNUM-1||r<0) {r=MAX_VARNUM-1;
    if (debug) current_token=TOKENIZER_ERROR;};
return r;
}
/*---------------------------------------------------------------------------*/
void tokenizer_string(char *dest, int len)
{
  char *string_end;
  int s,string_len;
  
 // if(tokenizer_token() != TOKENIZER_STRING) {return;}
  s=*ptr=='"';
  if (len==123) string_end = strpbrk(ptr + 1,"\r\n");
  else {string_end = strpbrk(ptr + 1,"\"\r\n");
	   string_end--;}
  if (string_end == NULL) {return;}
  string_len = string_end - ptr;
  if (len < string_len) string_len = len;
  memcpy(dest, ptr + s, string_len);
  dest[string_len] = 0;
  
  if (debug) {//short x=strlen(statement);
		//if (x-string_len<123)
	if (debug>1) strcat(statement,dest); else
		{short x=strlen(statement);
		statement[x]='$'; statement[x+1]='\0';}
	}
}
/*---------------------------------------------------------------------------*/
void tokenizer_statement(char *dest, int len)
{
  memcpy(dest, statement, len);
  short x=strlen(statement);
  if (x<len) len=x; 
  dest[len] = 0;
}
/*---------------------------------------------------------------------------*/
void tokenizer_label(char *dest, int len)
{
  char *string_end;
  int string_len;
  
  if (tokenizer_token() != TOKENIZER_LABEL) {
     return; }
  // allow string \r, \n, tabulation and space to end labels
  string_end = strpbrk(ptr + 1, " \t\r\n");

  if(string_end == NULL) {return;}
  string_len = string_end - ptr - 1;
  if (len < string_len) {string_len = len;}
  memcpy(dest, ptr + 1, string_len);
  dest[string_len] = 0;
  tokenizer_eol();
}
/*---------------------------------------------------------------------------*/
void tokenizer_error_print(void)
{
  DEBUG_PRINTF("tokenizer_error_print: '%s'\n", ptr);
}
/*---------------------------------------------------------------------------*/
int tokenizer_finished(void)
{
  return *ptr == 0 || current_token == TOKENIZER_ENDOFINPUT;
}
/*---------------------------------------------------------------------------*/
void tokenizer_end(void)
{
  first_line=1;
  current_line=1;
  nextptr=0;
  debug=0;
}